from .collection import (
    ApiCollection,
    ApiSearchCollection,
    Collection,
    CollectionExplorer,
    DbCollection,
    DbExplorer,
    ModelCollection,
    StaticCollection,
)
from .columns import Columns, DbColumns, TableColumns
from .data import (
    ApiData,
    ApiSearchData,
    BaseModelData,
    BaseSaData,
    CsvFileData,
    Data,
    DbData,
    ModelData,
    StatementModelData,
    StatementSaData,
    StaticData,
    TableData,
    UnionModelData,
    UnionSaData,
)
from .db_connection import (
    CkanDbConnection,
    DatastoreDbConnection,
    DbConnection,
    UrlDbConnection,
)
from .filters import Filters, TableFilters
from .pager import ClassicPager, OffsetPager, Pager, TemporalPager
from .serialize import (
    ChartJsSerializer,
    CsvSerializer,
    DictListSerializer,
    HtmlSerializer,
    HtmxTableSerializer,
    JsonlSerializer,
    JsonSerializer,
    Serializer,
    TableSerializer,
)

__all__ = [
    "CkanDbConnection",
    "DatastoreDbConnection",
    "DbCollection",
    "DbData",
    "TableData",
    "DbConnection",
    "UrlDbConnection",
    "ApiCollection",
    "ApiData",
    "ApiSearchCollection",
    "ApiSearchData",
    "BaseModelData",
    "BaseSaData",
    "ChartJsSerializer",
    "DictListSerializer",
    "ClassicPager",
    "TemporalPager",
    "OffsetPager",
    "Collection",
    "CollectionExplorer",
    "DbExplorer",
    "Columns",
    "DbColumns",
    "TableColumns",
    "CsvSerializer",
    "Data",
    "Filters",
    "TableFilters",
    "HtmlSerializer",
    "HtmxTableSerializer",
    "JsonSerializer",
    "JsonlSerializer",
    "ModelCollection",
    "ModelData",
    "Pager",
    "Serializer",
    "StatementModelData",
    "StatementSaData",
    "StaticCollection",
    "StaticData",
    "TableSerializer",
    "UnionModelData",
    "UnionSaData",
    "CsvFileData",
]



from __future__ import annotations

from typing import Any, Iterable

import sqlalchemy as sa

from ckanext.collection import internal, types


class Filters(
    types.BaseFilters,
    internal.Domain[types.TDataCollection],
):
    """Information about UI filters.

    Redefine its methods to produce information for UI filters and action
    buttions(Download, Export, etc.).

    """

    def make_filters(self) -> Iterable[types.Filter[Any]]:
        return []

    def make_actions(self) -> Iterable[types.Filter[Any]]:
        return []

    static_filters: Iterable[types.Filter[Any]] = internal.configurable_attribute(
        default_factory=lambda self: [],
    )
    static_actions: Iterable[types.Filter[Any]] = internal.configurable_attribute(
        default_factory=lambda self: [],
    )

    def __init__(self, obj: types.TDataCollection, /, **kwargs: Any):
        super().__init__(obj, **kwargs)
        self.filters = self.static_filters or self.make_filters()
        self.actions = self.static_actions or self.make_actions()


class DbFilters(Filters[types.TDbCollection]):
    pass


class TableFilters(DbFilters[types.TDbCollection]):
    table: str = internal.configurable_attribute()

    def make_filters(self) -> Iterable[types.Filter[Any]]:
        return [
            filter
            for c in self.attached.db_connection.inspector.get_columns(self.table)
            if (filter := self._filter_for(c))
        ]

    def _filter_for(self, column: dict[str, Any]) -> types.Filter[Any] | None:
        if column["name"] in self.attached.columns.filterable:
            if isinstance(column["type"], sa.Boolean):
                return self._boolean_filter(column)

            return self._input_filter(column)

        if column["name"] in self.attached.columns.searchable:
            return self._input_filter(column)

        return None

    def _filter_label(self, column: dict[str, Any]) -> str:
        return self.attached.columns.labels.get(column["name"], column["name"])

    def _boolean_filter(self, column: dict[str, Any]) -> types.Filter[Any]:
        options = [
            types.SelectOption(text="All", value=""),
            types.SelectOption(text="Yes", value="t"),
            types.SelectOption(text="No", value="f"),
        ]
        return types.SelectFilter(
            name=column["name"],
            type="select",
            options=types.SelectFilterOptions(
                label=self._filter_label(column),
                options=options,
            ),
        )

    def _distinct_filter(self, column: dict[str, Any]) -> types.Filter[Any]:
        with self.attached.db_connection.engine.connect() as conn:
            values = conn.scalars(
                sa.select(sa.distinct(sa.column(column["name"]))).select_from(
                    sa.table(self.table),
                ),
            )
        options = [types.SelectOption(text="All", value="")] + [
            types.SelectOption(text=v, value=v) for v in values
        ]
        return types.SelectFilter(
            name=column["name"],
            type="select",
            options=types.SelectFilterOptions(
                label=self._filter_label(column),
                options=options,
            ),
        )

    def _input_filter(self, column: dict[str, Any]) -> types.Filter[Any]:
        return types.InputFilter(
            name=column["name"],
            type="input",
            options=types.InputFilterOptions(
                label=self._filter_label(column),
            ),
        )



"""Logic used across collection utilities."""

from __future__ import annotations

import abc
import dataclasses
import inspect
import logging
from collections.abc import Hashable
from typing import Any, Callable, Generic, TypeVar, cast

import ckan.plugins.toolkit as tk

from . import types

log = logging.getLogger(__name__)
T = TypeVar("T")


class Sentinel: ...


SENTINEL: Any = Sentinel()


@dataclasses.dataclass
class Registry(Generic[T]):
    members: dict[Hashable, T]

    def reset(self):
        self.members.clear()

    def register(self, name: Hashable, member: T):
        self.members[name] = member

    def get(self, name: Hashable) -> T | None:
        return self.members.get(name)


collection_registry: Registry[types.CollectionFactory] = Registry({})


class AttachTrait(abc.ABC, Generic[types.TDataCollection]):
    """Attach collection to the current object.

    `_attach` method must be called as early as possible in the constructor of
    the derived class. It makes collection available inside an instance via
    `attached` property. Because initialization of collection utilities often
    relies on collection details, you should call `_attach` before any other
    logic.

    Example:
    >>> class Impl(AttachTrait):
    >>>     def __init__(self, collection):
    >>>         self._attach(collection)
    >>>
    >>> collection = object()
    >>> obj = Impl(collection)
    >>> assert obj.attached is collection

    """

    __collection: types.TDataCollection

    def _attach(self, obj: types.TDataCollection):
        self.__collection = obj
        if isinstance(self, types.Service):
            # this block allows attaching services to non-collections. Mainly
            # it's used in tests
            replace_service = getattr(obj, "replace_service", None)
            if replace_service:
                replace_service(self)

    @property
    def attached(self) -> types.TDataCollection:
        return self.__collection


class AttrSettingsTrait:
    """Get attribute value from settings for the current utility object.

    To mark an attribute as configurable, declare it using
    `configurable_attribute` function:

    >>> class Impl(AttrSettingsTrait):
    >>>     attr = configurable_attribute("default value")

    To initialize attributes, call `_gather_settings` with dictionary
    containing attribute values. If attribute is available in the dictionary,
    dictionary value moved into attribute:

    >>> obj = Impl()
    >>> obj._gather_settings({"attr": "custom value"})
    >>> assert obj.attr == "custom value"

    If dictionary doesn't have a member named after the attribute, default
    value/factory is used to initialize the attribute

    >>> obj = Impl()
    >>> obj._gather_settings({})
    >>> assert obj.attr == "default value"

    `_gather_settings` can be called multiple times to reset/modify
    settings. Keep in mind, that every call processes all the configurable
    attributes of the instance. For example, if instance has `a` and `b`
    configurable attributes and you call `_gather_settings` with `{"a": 1}`
    first, and then with `{"b": 2}`, value of `a` will be removed and replaced
    by the default:

    >>> class Impl(AttrSettingsTrait):
    >>>     a = configurable_attribute(None)
    >>>     b = configurable_attribute(None)
    >>>
    >>> obj = Impl()
    >>> obj._gather_settings({"a": 1})
    >>> assert obj.a == 1
    >>> assert obj.b is None
    >>>
    >>> obj._gather_settings({"b": 2})
    >>> assert obj.a == None
    >>> assert obj.b is 2
    >>>
    >>> obj._gather_settings({"b": 2, "a": 1})
    >>> assert obj.a == 1
    >>> assert obj.b == 2
    >>>
    >>> obj._gather_settings({})
    >>> assert obj.a is None
    >>> assert obj.b is None

    The best place to gather settings, early in the constructor, right after
    `_attach`. Example:

    >>> class Impl(AttachTrait, AttrSettingsTrait):
    >>>     a = configurable_attribute(None)
    >>>     b = configurable_attribute(None)
    >>>
    >>>     def __init__(self, collection, **kwargs):
    >>>         self._attach(collection)
    >>>         self._gather_settings(kwargs)

    """

    def _gather_settings(self, kwargs: dict[str, Any]):
        for k, attr in inspect.getmembers(
            type(self),
            lambda attr: isinstance(attr, _InitAttr),
        ):
            attr: _InitAttr
            if k in kwargs:
                setattr(self, k, kwargs.pop(k))

            else:
                setattr(self, k, attr.get_default(self, k))


@dataclasses.dataclass
class _InitAttr:
    default: Any = SENTINEL
    default_factory: Callable[[Any], Any] = SENTINEL

    def __post_init__(self):
        if self.default is not self.default_factory:
            return

    def get_default(self, obj: Any, name: str):
        if self.default is not SENTINEL:
            return self.default

        if self.default_factory is not SENTINEL:
            return self.default_factory(obj)

        msg = (
            f"__init__() of {obj.__class__.__name__} missing "
            + f"1 required keyword-only argument: '{name}'"
        )
        raise TypeError(msg)


def configurable_attribute(
    default: T | Sentinel = SENTINEL,
    default_factory: Callable[[Any], T] = SENTINEL,
) -> T:
    """Declare configurable attribute.

    Example:
    >>> class DataFactory(Data):
    >>>     private = configurable_attribute(False)
    >>>
    >>> data = DataFactory(None, private=True)
    >>> assert data.private
    """
    return cast(T, _InitAttr(default, cast(Any, default_factory)))


class UserTrait(AttrSettingsTrait):
    """Add configurable `user` attribute.

    Default value set to `tk.current_user.name`.
    """

    user = configurable_attribute(
        default_factory=lambda self: tk.current_user.name if tk.current_user else "",
    )


class Domain(AttachTrait[types.TDataCollection], AttrSettingsTrait):
    """Standard initializer for collection utilities.

    Used as base class for utility instances created during collection
    initialization(e.g Pager, Columns, Filters, Data).

    Defines standard constructor signature, attaches collection to the utility
    instance and collects settings.

    """

    def __init__(self, obj: types.TDataCollection, /, **kwargs: Any):
        self._attach(obj)
        self._gather_settings(kwargs)

    @classmethod
    def with_attributes(
        cls: type[T],
        name: str | None = None,
        /,
        **attributes: Any,
    ) -> type[T]:
        """Create anonymous derivable of the class with overriden attributes.

        This is a shortcut for defining a proper subclass of the domain:

        >>> class Parent(Domain):
        >>>     prop = "parent value"
        >>>
        >>> class Child(Parent):
        >>>     prop = "child value"
        >>>     custom = "custom value"
        >>>
        >>> # equivalent
        >>> Child = Parent.with_attributes(prop="child value", custom="custom value")
        """
        return cast("type[T]", type(name or cls.__name__, (cls,), attributes))


def parse_sort(sort: str) -> tuple[str, bool]:
    """Parse sort as column and sorting direction.

    Turn `-name` and `name desc` into `(name, True)`. Everything else turns
    into `(name, False)`.

    """
    desc = False
    if sort.startswith("-"):
        sort = sort[1:]
        desc = True

    elif len(parts := sort.split()) == 2:
        sort, direction = parts

        if direction.lower() == "desc":
            desc = True

    return sort, desc


def get_collection(
    name: str,
    params: dict[str, Any],
    prefix_params: bool = False,
    /,
    **kwargs: Any,
) -> types.BaseCollection | None:
    if prefix_params:
        params = {f"{name}:{k}": v for k, v in params.items()}
    if factory := collection_registry.get(name):
        return factory(name, params, **kwargs)

    return None



from __future__ import annotations

import csv
import logging

from ckanext.collection import internal, types

from .base import Data

log = logging.getLogger(__name__)


class CsvFileData(Data[types.TData, types.TDataCollection]):
    """Data source for CSV file.

    CSV file available at path specified by `source` attribute of the service
    is read into memory and its every row transformed into dictionary.

    Warning:
        Iteration and size measurement uses cached version of source's content.
        If `source` attribute was overriden or its content was modified after
        service initialization, call `refresh_data()` method
        of the service to reset the cache.

    Attributes:
        source: path to CSV source

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     data_factory=data.CsvFileData,
        >>>     data_settings={"source": "/path/to/file.csv"},
        >>> )
        >>> list(col)
        [
            {"column_1": "value_1", "column_2": "value_2"},
            ...
        ]
        ```



    """

    source: str = internal.configurable_attribute()

    def compute_data(self):
        with open(self.source) as src:
            reader = csv.DictReader(src)
            return list(reader)



from __future__ import annotations

import csv
import json
import os
from io import StringIO
from typing import Any, cast

import pytest

from ckan import model
from ckan.tests.factories import CKANFactory

import ckanext.collection.utils as cu


class TestModelCollection(cu.Collection):
    DataFactory = cu.ModelData.with_attributes(model=model.Resource)
    ColumnsFactory = cu.Columns.with_attributes(names=["name", "size"])


## collection of all packages available via search API
class TestApiCollection(cu.Collection):
    DataFactory = cu.ApiSearchData.with_attributes(action="package_search")
    ColumnsFactory = cu.Columns.with_attributes(names=["name", "title"])


## collection of all records from CSV file
class TestCsvCollection(cu.Collection):
    DataFactory = cu.CsvFileData.with_attributes(
        source=os.path.join(os.path.dirname(__file__), "data/file.csv"),
    )
    ColumnsFactory = cu.Columns.with_attributes(names=["a", "b"])


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_index")
class TestQuickstart:
    def test_model(self, resource_factory: type[CKANFactory]):
        col = TestModelCollection("", {})
        assert col.data.total == 0

        resource_factory.create_batch(3)

        col = TestModelCollection("", {})
        assert col.data.total == 3

        serializer = cu.JsonSerializer(col)
        expected = [{"name": r.name, "size": r.size} for r in col.data]
        assert json.loads(serializer.serialize()) == expected

        serializer = cu.CsvSerializer(col)
        output = StringIO()
        writer = csv.writer(output)
        writer.writerow(["name", "size"])
        writer.writerows([[r.name, r.size] for r in col.data])
        assert serializer.serialize() == output.getvalue()

        serializer = cu.DictListSerializer(col)
        expected = [{"name": r.name, "size": r.size} for r in col.data]
        assert serializer.serialize() == expected

    def test_api(self, package_factory: type[CKANFactory]):
        col = TestApiCollection("", {})
        assert col.data.total == 0

        package_factory.create_batch(3)

        col = TestApiCollection("", {})
        assert col.data.total == 3

        serializer = cu.JsonSerializer(col)
        expected = [{"name": r["name"], "title": r["title"]} for r in col.data]
        assert json.loads(serializer.serialize()) == expected

        serializer = cu.CsvSerializer(col)
        output = StringIO()
        writer = csv.writer(output)
        writer.writerow(["name", "title"])
        writer.writerows([[r["name"], r["title"]] for r in col.data])
        assert serializer.serialize() == output.getvalue()

        serializer = cu.DictListSerializer(col)
        expected = [{"name": r["name"], "title": r["title"]} for r in col.data]
        assert serializer.serialize() == expected

    def test_csv(self, package_factory: type[CKANFactory]):
        filename = cast(Any, TestCsvCollection.DataFactory).source
        col = TestCsvCollection("", {})
        assert col.data.total == 16

        assert len(list(col)) == 10
        assert len(list(col.data)) == 16
        assert len(list(col.data.range(2, 5))) == 3

        serializer = cu.JsonSerializer(col)
        with open(filename) as src:
            expected = [{"a": r["a"], "b": r["b"]} for r in csv.DictReader(src)]
        assert json.loads(serializer.serialize()) == expected

        serializer = cu.CsvSerializer(col)
        output = StringIO()
        with open(filename) as src:
            writer = csv.writer(output)
            for row in csv.reader(src):
                writer.writerow(row[:-1])
        assert serializer.serialize() == output.getvalue()

        serializer = cu.DictListSerializer(col)
        with open(filename) as src:
            expected = [{"a": r["a"], "b": r["b"]} for (r) in csv.DictReader(src)]
        assert serializer.serialize() == expected



from __future__ import annotations

import copy
import logging
from typing import Any, Callable, Iterable, Iterator

import ckan.plugins.toolkit as tk
from ckan.types import Context

from ckanext.collection import internal, types

from .base import Data

log = logging.getLogger(__name__)


class ApiData(Data[types.TData, types.TDataCollection], internal.UserTrait):
    """API data source.

    This base class is suitable for building API calls. Its `compute_data`
    makes the single request to the specified API action and yields items from
    the response.

    Warning:
        Iteration and size measurement uses cached version of action's result.
        To call action again and use the fresh result, use `refresh_data()` method
        of the service.

    Attributes:
        action: API action that returns the data
        payload: parameters passed to the action
        ignore_auth: skip authorization checks
        user (str): name of the user for the action. Default: `tk.current_user.name`

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     data_factory=data.ApiData,
        >>>     data_settings={"action": "group_list_authz", "user": "default"},
        >>> )
        >>> list(col)
        [{...}, {...}]
        ```

    """

    action: str = internal.configurable_attribute()
    payload: dict[str, Any] = internal.configurable_attribute(
        default_factory=lambda self: {},
    )
    ignore_auth: bool = internal.configurable_attribute(False)

    def make_context(self):
        return Context(user=self.user, ignore_auth=self.ignore_auth)

    def prepare_payload(self) -> dict[str, Any]:
        return copy.deepcopy(self.payload)

    def compute_data(self):
        action = tk.get_action(self.action)
        return action(self.make_context(), self.prepare_payload())


class ApiSearchData(ApiData[types.TData, types.TDataCollection]):
    """API data source optimized for package_search-like actions.

    This class expects that API action accepts `start` and `rows` parameters
    that controls offset and limit. And result of the action must contain
    `count` and `results` keys.

    This data service can iterate over huge number of items, reading just few
    of them into the memory at once. It means you can iterate over thousands of
    datasets in an efficient way without messing with offsets and limits:

    ```python
    packages = collection.Collection(
        data_factory=data.ApiSearchData,
        data_settings={"action": "package_search"},
    )

    # the following loop goes over every available package, no matter how may
    # of them are available inside search index
    for pkg in packages.data: # (1)!
        ...

    # total number of items processed in the previous step is the same as
    # number of public datasets
    assert packages.data.total == tk.get_action("package_search")({}, {})["count"]
    ```

    1. `data` service is iterated directly to access all the items. Iterating
       over `packages` would yield just first 10 items because of pagination.


    Warning:
        Size measurement uses cached number of records. To refresh the total
        number of records, call `refresh_data()` method of the service. Such
        thing can be used if a new item was created after initialization of the
        collection

        The records are not cached, to reduce memory usage. Every separate
        iteration over data service and every `range` call initiates a fresh
        API request.

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     data_factory=data.ApiSearchData,
        >>>     data_settings={
        >>>         "action": "package_search",
        >>>         "payload": {"q": "res_format:CSV"},
        >>>     },
        >>> )
        >>> list(col)
        [{...}, {...}]
        ```

    """

    start_param = internal.configurable_attribute("start")
    rows_param = internal.configurable_attribute("rows")

    def prepare_payload(self) -> dict[str, Any]:
        payload = super().prepare_payload()
        return dict(
            payload,
            **self.get_filters(),
            **self.get_sort(),
        )

    def get_filters(self) -> dict[str, str]:
        return {}

    def get_sort(self) -> dict[str, str]:
        sort = self.attached.params.get("sort")
        if not sort:
            return {}

        column, desc = internal.parse_sort(sort)

        if column not in self.attached.columns.sortable:
            log.warning("Unexpected sort value: %s", sort)
            return {}

        direction = "desc" if desc else "asc"
        return {"sort": f"{column} {direction}"}

    def get_action(self) -> Callable[[Context, dict[str, Any]], Any]:
        return tk.get_action(self.action)

    def compute_data(self):
        action = self.get_action()
        payload = self.prepare_payload()
        payload[self.rows_param] = 0

        return action(self.make_context(), payload)

    def compute_total(self, data: dict[str, Any]) -> int:
        return data["count"]

    def range(self, start: int, end: int) -> Iterable[types.TData]:
        action = self.get_action()

        payload = self.prepare_payload()
        payload[self.rows_param] = end - start
        payload[self.start_param] = start

        return action(self.make_context(), payload)["results"]

    def at(self, index: int) -> types.TData:
        action = self.get_action()
        payload = self.prepare_payload()
        payload[self.rows_param] = 1
        payload[self.start_param] = index

        return action(self.make_context(), payload)["results"][0]

    def __iter__(self) -> Iterator[types.TData]:
        action = self.get_action()
        context = self.make_context()
        start = 0
        payload = self.prepare_payload()
        payload[self.start_param] = start

        while True:
            result = action(context, payload)

            yield from result["results"]
            start += len(result["results"])

            if start >= result["count"] or not result["results"]:
                break



from __future__ import annotations

import abc
from typing import Any, Iterable

import sqlalchemy as sa
from sqlalchemy.sql.selectable import GenerativeSelect, Select

from ckanext.collection import internal, types

from .model import BaseSaData

# from pandas.compat._optional import import_optional_dependency
# import_optional_dependency("sqlalchemy", errors="ignore")


class DbData(BaseSaData[Select, types.TData, types.TDbCollection], abc.ABC):
    def _execute(self, stmt: GenerativeSelect):
        return self.attached.db_connection.engine.execute(stmt)


class TableData(DbData[types.TData, types.TDbCollection]):
    table: str = internal.configurable_attribute()
    static_columns: Iterable[Any] = internal.configurable_attribute(
        default_factory=lambda self: [],
    )

    def get_base_statement(self):
        columns = self.static_columns or [
            sa.column(c["name"])
            for c in self.attached.db_connection.inspector.get_columns(self.table)
        ]
        return sa.select(*columns).select_from(
            sa.table(self.table),
        )



from __future__ import annotations

import pytest

from ckanext.collection.utils import Collection, pager


@pytest.fixture()
def collection():
    return Collection("", {})


class TestClassicPager:
    def test_settings(self, collection: Collection):
        obj = pager.ClassicPager(collection)
        assert obj.page == 1
        assert obj.size == 10

        obj = pager.ClassicPager(collection, page=5, rows_per_page=101)
        assert obj.page == 5
        assert obj.size == 101

    @pytest.mark.parametrize(
        ("page", "size", "start", "end"),
        [
            (1, 10, 0, 10),
            (3, 2, 4, 6),
            (10, 1, 9, 10),
        ],
    )
    def test_logic(
        self,
        collection: Collection,
        page: int,
        size: int,
        start: int,
        end: int,
    ):
        obj = pager.ClassicPager(
            collection,
            page=page,
            rows_per_page=size,
        )
        assert obj.size == size
        assert obj.start == start
        assert obj.end == end



from __future__ import annotations

from . import types
from .interfaces import ICollection
from .internal import (
    Domain,
    UserTrait,
    configurable_attribute,
    get_collection,
    parse_sort,
)
from .utils import collection, columns, data, db_connection, filters, pager, serialize

__all__ = [
    "get_collection",
    "parse_sort",
    "Domain",
    "UserTrait",
    "configurable_attribute",
    "types",
    "ICollection",
    "collection",
    "columns",
    "data",
    "db_connection",
    "filters",
    "pager",
    "serialize",
]



from __future__ import annotations

import abc
import csv
import io
import itertools
import json
import operator
from functools import reduce
from typing import Any, Callable, Generic, Iterable, cast

import sqlalchemy as sa
from sqlalchemy.engine import Row
from sqlalchemy.exc import NoInspectionAvailable
from sqlalchemy.orm import InstanceState

import ckan.plugins.toolkit as tk

from ckanext.collection import internal, types

__all__ = [
    "Serializer",
    "StreamingSerializer",
    "RenderableSerializer",
    "CsvSerializer",
    "DictListSerializer",
    "JsonlSerializer",
    "JsonSerializer",
    "ChartJsSerializer",
    "HtmlSerializer",
    "TableSerializer",
    "HtmxTableSerializer",
]


def basic_row_dictizer(row: Any) -> dict[str, Any]:
    if isinstance(row, dict):
        return cast("dict[str, Any]", row)

    if isinstance(row, Row):
        if hasattr(row, "_asdict"):
            return row._asdict()  # # type: ignore

        if hasattr(row, "keys"):
            return dict(zip(row.keys(), row))

    try:
        reflection = sa.inspect(row)
    except NoInspectionAvailable:
        raise TypeError(type(row)) from None

    if isinstance(reflection, InstanceState):
        return {attr.key: attr.value for attr in reflection.attrs}

    raise TypeError(type(row))


class Serializer(
    types.BaseSerializer,
    internal.Domain[types.TDataCollection],
    Generic[types.TSerialized, types.TDataCollection],
):
    """Base collection serializer.

    For any derived implementation, `serialize` must transfrom data of the
    collection into expected format.

    Example:
        ```pycon
        >>> class MySerializer(serialize.Serializer):
        >>>     def stream(self):
        >>>         for record in self.attached.data:
        >>>             yield yaml.dump(record)
        >>>
        >>>     def serialize(self):
        >>>         return "---\\n".join(self.stream())
        ```

    """

    value_serializers: dict[
        str, types.ValueSerializer
    ] = internal.configurable_attribute(
        default_factory=lambda self: {},
    )
    row_dictizer: Callable[[Any], dict[str, Any]] = internal.configurable_attribute(
        basic_row_dictizer,
    )

    def serialize(self) -> types.TSerialized:
        """Return serialized data."""
        result: Any = (r for r in self.attached.data)
        return result

    def serialize_value(self, value: Any, name: str, record: Any):
        """Transform record's value into its serialized form."""
        for serializer, options in self.attached.columns.serializers.get(name, []):
            if isinstance(serializer, str):
                serializer = self.value_serializers.get(serializer)

            if not serializer:
                continue

            value = serializer(value, options, name, record, self)

        return value

    def dictize_row(self, row: Any) -> dict[str, Any]:
        """Transform single data record into serializable dictionary."""
        result = self.row_dictizer(row)
        if fields := self.attached.columns.names:
            visible = self.attached.columns.visible
        else:
            fields = list(result)
            visible = set(fields)

        return {
            field: self.serialize_value(result.get(field), field, row)
            for field in fields
            if field in visible
        }


class StreamingSerializer(
    Serializer[types.TSerialized, types.TDataCollection],
):
    @abc.abstractmethod
    def stream(self) -> Iterable[types.TSerialized]:
        """Iterate over fragments of the content.

        Type of the stream fragment must be the same as type of serialized
        content. For example, serializer that produces list of dictionaries,
        must yield `[dict(...)]`, not just `dict(...)`
        """
        raise NotImplementedError

    def serialize(self) -> types.TSerialized:
        return reduce(operator.add, self.stream())


class DictListSerializer(
    StreamingSerializer["list[dict[str, Any]]", types.TDataCollection],
):
    """Convert data into a list of dictionaries.

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     serializer_factory=serialize.DictListSerializer,
        >>>     columns_factory=columns.Columns.with_attributes(
        >>>         names=["name", "sysadmin"],
        >>>     ),
        >>>     data_factory=data.ModelData,
        >>>     data_settings={"model": model.User},
        >>> )
        >>> col.serializer.serialize()
        [{'name': 'default', 'sysadmin': True}]
        ```

    """

    def stream(self):
        """Iterate over fragments of the content."""
        for item in self.attached.data:
            yield [self.dictize_row(item)]


class RenderableSerializer(StreamingSerializer[str, types.TDataCollection]):
    def stream(self) -> Iterable[str]:
        """Iterate over fragments of the content."""
        yield ""

    def render(self) -> str:
        """Combine content fragments into a single dump."""
        return self.serialize()


class CsvSerializer(StreamingSerializer[str, types.TDataCollection]):
    """Serialize collection into CSV document.

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     serializer_factory=serialize.CsvSerializer,
        >>>     columns_factory=columns.Columns.with_attributes(
        >>>         names=["name", "sysadmin"],
        >>>     ),
        >>>     data_factory=data.ModelData,
        >>>     data_settings={"model": model.User},
        >>> )
        >>> col.serializer.serialize()
        name,sysadmin
        default,True
        ```
    """

    def get_writer(self, buff: io.StringIO, fieldnames: list[str]):
        return csv.DictWriter(buff, fieldnames, extrasaction="ignore")

    def get_header_row(self, writer: csv.DictWriter[str]) -> dict[str, str]:
        return {
            col: self.attached.columns.labels.get(col, col) for col in writer.fieldnames
        }

    def prepare_row(self, row: Any) -> dict[str, Any]:
        return self.dictize_row(row)

    def stream(self) -> Iterable[str]:
        buff = io.StringIO()
        data = iter(self.attached.data)
        fieldnames = [
            n for n in self.attached.columns.names if n in self.attached.columns.names
        ]
        if not fieldnames:
            empty = object()
            row = next(data, empty)
            if row is not empty:
                data = itertools.chain([row], data)
                fieldnames = list(self.prepare_row(row))

        writer = self.get_writer(buff, fieldnames)

        writer.writerow(self.get_header_row(writer))

        yield buff.getvalue()
        buff.seek(0)
        buff.truncate()

        for row in data:
            writer.writerow(self.prepare_row(row))
            yield buff.getvalue()
            buff.seek(0)
            buff.truncate()


class JsonlSerializer(StreamingSerializer[str, types.TDataCollection]):
    """Serialize collection into JSONL lines.

    Attributes:
        encoder: `json.JSONEncoder` instance used for serialization

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     serializer_factory=serialize.JsonlSerializer,
        >>>     columns_factory=columns.Columns.with_attributes(
        >>>         names=["name", "sysadmin"],
        >>>     ),
        >>>     data_factory=data.ModelData,
        >>>     data_settings={"model": model.User},
        >>> )
        >>> col.serializer.serialize()
        {"name: "default", "sysadmin": true}
        {"name: "normal_user", "sysadmin": false}
        ```
    """

    encoder = internal.configurable_attribute(json.JSONEncoder(default=str))

    def stream(self) -> Iterable[str]:
        buff = io.StringIO()
        for row in map(self.dictize_row, self.attached.data):
            buff.write(self.encoder.encode(row))
            yield buff.getvalue()
            yield "\n"
            buff.seek(0)
            buff.truncate()


class JsonSerializer(StreamingSerializer[str, types.TDataCollection]):
    """Serialize collection into single JSON document.

    Attributes:
        encoder: `json.JSONEncoder` instance used for serialization

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     serializer_factory=serialize.JsonSerializer,
        >>>     columns_factory=columns.Columns.with_attributes(
        >>>         names=["name", "sysadmin"],
        >>>     ),
        >>>     data_factory=data.ModelData,
        >>>     data_settings={"model": model.User},
        >>> )
        >>> col.serializer.serialize()
        [{"name: "default", "sysadmin": true},
         {"name: "normal_user", "sysadmin": false}]
        ```
    """

    encoder = internal.configurable_attribute(json.JSONEncoder(default=str))

    def stream(self):
        yield self.encoder.encode(
            [self.dictize_row(row) for row in self.attached.data],
        )


class ChartJsSerializer(StreamingSerializer[str, types.TDataCollection]):
    """Serialize collection into data source for ChartJS."""

    label_column: str = internal.configurable_attribute("")
    dataset_columns: list[str] = internal.configurable_attribute(
        default_factory=lambda self: [],
    )
    dataset_labels: dict[str, str] = internal.configurable_attribute(
        default_factory=lambda self: {},
    )
    colors: dict[str, str] = internal.configurable_attribute(
        default_factory=lambda self: {},
    )

    def stream(self):
        labels = []
        data: list[list[int]] = []

        for item in map(self.dictize_row, self.attached.data):
            labels.append(item.get(self.label_column, None))
            data.append(
                [item[name] for name in self.dataset_columns],
            )

        datasets: list[dict[str, Any]] = [
            {
                "data": row,
                "label": self.dataset_labels.get(column, column),
                "backgroundColor": self.colors.get(column),
            }
            for column, row in zip(self.dataset_columns, zip(*data))
        ]
        yield json.dumps(
            {
                "labels": labels,
                "datasets": datasets,
            },
        )


class HtmlSerializer(RenderableSerializer[types.TDataCollection]):
    """Serialize collection into HTML document.

    Attributes:
        main_template: path to Jinja2 template for the collection
        record_template: path to Jinja2 template for the individual record

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     serializer_factory=serialize.HtmlSerializer,
        >>>     data_factory=data.StaticData,
        >>>     data_settings={"data": range(3)},
        >>> )
        >>> col.serializer.serialize()
        <ul>
          <li>0</li>
          <li>1</li>
          <li>2</li>
        </ul>
        ```
    """

    ensure_dictized: bool = internal.configurable_attribute(False)

    main_template: str = internal.configurable_attribute(
        "collection/serialize/html/main.html",
    )
    record_template: str = internal.configurable_attribute(
        "collection/serialize/html/record.html",
    )

    prefix: str = internal.configurable_attribute("collection-content")
    base_class: str = internal.configurable_attribute("collection")

    @property
    def form_id(self):
        return f"{self.prefix}-form--{self.attached.name}"

    def get_data(self) -> dict[str, Any]:
        return {
            "collection": self.attached,
        }

    def stream(self):
        import contextlib

        from flask import has_app_context

        from ckan.lib.helpers import _get_auto_flask_context  # type: ignore

        ctx: Any = (
            contextlib.nullcontext() if has_app_context() else _get_auto_flask_context()
        )
        with ctx:
            yield tk.render(
                self.main_template,
                self.get_data(),
            )


class TableSerializer(HtmlSerializer[types.TDataCollection]):
    """Serialize collection into HTML table.

    Attributes:
        main_template: path to template for the collection
        main_template: path to template for the table wrapper
        record_template: path to template for the individual record
        counter_template: path to template for the item counter
        pager_template: path to template for the table pagination widget
        form_template: path to template for the search form
        filter_template: path to template for filters block inside search form

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     serializer_factory=serialize.TableSerializer,
        >>>     columns_factory=columns.Columns.with_attributes(
        >>>         names=["name", "sysadmin"],
        >>>     ),
        >>>     data_factory=data.ModelData,
        >>>     data_settings={"model": model.User},
        >>> )
        >>> col.serializer.serialize()
        <table>...</table>
        ```
    """

    main_template: str = internal.configurable_attribute(
        "collection/serialize/table/main.html",
    )
    table_template: str = internal.configurable_attribute(
        "collection/serialize/table/table.html",
    )
    record_template: str = internal.configurable_attribute(
        "collection/serialize/table/record.html",
    )
    counter_template: str = internal.configurable_attribute(
        "collection/serialize/table/counter.html",
    )
    pager_template: str = internal.configurable_attribute(
        "collection/serialize/table/pager.html",
    )
    form_template: str = internal.configurable_attribute(
        "collection/serialize/table/form.html",
    )
    filter_template: str = internal.configurable_attribute(
        "collection/serialize/table/filter.html",
    )

    prefix: str = internal.configurable_attribute("collection-table")

    @property
    def table_id(self):
        return f"{self.prefix}-id--{self.attached.name}"


class HtmxTableSerializer(TableSerializer[types.TDataCollection]):
    """Serialize collection into HTML table powered by HTMX.

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     serializer_factory=serialize.HtmxTableSerializer,
        >>>     columns_factory=columns.Columns.with_attributes(
        >>>         names=["name", "sysadmin"],
        >>>     ),
        >>>     data_factory=data.ModelData,
        >>>     data_settings={"model": model.User},
        >>> )
        >>> col.serializer.serialize()
        <table>...</table>
        ```
    """

    main_template: str = internal.configurable_attribute(
        "collection/serialize/htmx_table/main.html",
    )
    table_template: str = internal.configurable_attribute(
        "collection/serialize/htmx_table/table.html",
    )
    record_template: str = internal.configurable_attribute(
        "collection/serialize/htmx_table/record.html",
    )
    counter_template: str = internal.configurable_attribute(
        "collection/serialize/htmx_table/counter.html",
    )
    pager_template: str = internal.configurable_attribute(
        "collection/serialize/htmx_table/pager.html",
    )
    form_template: str = internal.configurable_attribute(
        "collection/serialize/htmx_table/form.html",
    )
    filter_template: str = internal.configurable_attribute(
        "collection/serialize/htmx_table/filter.html",
    )

    debug: bool = internal.configurable_attribute(False)
    push_url: bool = internal.configurable_attribute(False)
    base_class: str = internal.configurable_attribute("htmx-collection")

    render_url: str = internal.configurable_attribute(
        default_factory=lambda self: tk.h.url_for(
            "ckanext-collection.render",
            name=self.attached.name,
        ),
    )



from __future__ import annotations

import click

from ckanext.collection.internal import collection_registry

__all__ = ["collection"]


@click.group(short_help="ckanext-collection CLI")
def collection():
    pass


@collection.command("list")
@click.option("--name-only", is_flag=True, help="Show only collection names")
def list_collections(name_only: bool):
    """List all registered collections."""
    for name, collection in collection_registry.members.items():
        line = name if name_only else f"{name}: {collection}"
        click.secho(line)



from __future__ import annotations

import logging
from typing import Iterable

from ckanext.collection import internal, types

from .api import ApiData, ApiSearchData
from .base import Data
from .db import DbData, TableData
from .misc import CsvFileData
from .model import BaseSaData, ModelData, StatementSaData, TemporalSaData, UnionSaData

__all__ = [
    "Data",
    "CsvFileData",
    "TableData",
    "ApiData",
    "ApiSearchData",
    "UnionSaData",
    "UnionModelData",
    "StatementSaData",
    "StatementModelData",
    "BaseSaData",
    "BaseModelData",
    "DbData",
    "ModelData",
    "TemporalSaData",
]

BaseModelData = BaseSaData
UnionModelData = UnionSaData
StatementModelData = StatementSaData

log = logging.getLogger(__name__)


class StaticData(Data[types.TData, types.TDataCollection]):
    """Static data source.

    This class produce items from its `data` attribute. Use any sequence as a
    value for `data` during initialization.

    Warning:
        Iteration and size measurement uses cached version of `data`. If `data`
        attribute was overriden after service initialization, call
        `refresh_data()` method of the service to reset the cache.

    Attributes:
        data: sequence of items produced by the service

    Example:
        ```python
        NumericData = data.StaticData.with_attributes(data=range(1, 20))

        UppercaseData = data.StaticData.with_attributes(
            data="ABCDEFGHIJKLMNOPQRSTUVWXYZ",
        )
        ```
        ```pycon
        >>> col = collection.Collection(data_factory=NumericData)
        >>> list(col)
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        ```

    """

    data: Iterable[types.TData] = internal.configurable_attribute(
        default_factory=lambda self: [],
    )

    def compute_data(self) -> Iterable[types.TData]:
        return self.data



from __future__ import annotations

from typing import Any

import ckan.plugins.toolkit as tk
from ckan import types
from ckan.logic import authz

from ckanext.collection import config


@tk.auth_allow_anonymous_access
def collection_view_render(context: types.Context, data_dict: dict[str, Any]):
    if data_dict["name"] in config.anonymous_collections():
        return {"success": True}

    if context["user"] and data_dict["name"] in config.authenticated_collections():
        return {"success": True}

    return {"success": False}


@tk.auth_allow_anonymous_access
def collection_view_export(context: types.Context, data_dict: dict[str, Any]):
    return authz.is_authorized("collection_view_render", context, data_dict)



from __future__ import annotations

import abc
import logging
from datetime import date
from functools import cached_property
from typing import Any, Generic, Iterable, Iterator, TypeVar, cast

import sqlalchemy as sa
from sqlalchemy.orm import Mapper
from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql.elements import Label
from sqlalchemy.sql.selectable import GenerativeSelect, Select

from ckan import model
from ckan.types import AlchemySession

from ckanext.collection import internal, types

from .base import Data

log = logging.getLogger(__name__)
TStatement = TypeVar("TStatement", bound=GenerativeSelect)


class BaseSaData(
    Data[types.TData, types.TDataCollection],
    Generic[TStatement, types.TData, types.TDataCollection],
):
    """Abstract data source for SQL statements.

    This class can be extended to build data source over SQL statement. Its
    `compute_data` calls 4 methods:

    * `get_base_statement`: produces initial statement
    * `alter_statement(stmt)`: modifies statement or replaces it completely
    * `statement_with_filters(stmt)`: apply `WHERE` and `HAVING` conditions
    * `statement_with_sorting(stmt)`: apply `ORDER BY`

    These methods do nothing by default, but can be replaced in sublcasses to
    build SQL statement gradually.

    Warning:
        Final statement produced by `compute_data` and total number of results
        are cached. Call `refresh_data()` method of the service to rebuild the
        statement and refresh number of rows.

    Attributes:
        use_naive_filters: search by filterable columns from `params`. Default: true
        use_naive_search: if `params` contains `q`, ILIKE it against searchable
            columns. Default: true
        session: SQLAlchemy session
        is_scalar: return only first column from each row

    Example:
        ```python
        import sqlalchemy as sa
        from ckan import model

        class UserData(data.BaseSaData):
            def get_base_statement(self):
                return sa.select(model.User.name)
        ```

        ```pycon
        >>> col = collection.Collection(data_factory=UserData)
        >>> list(col)
        [("default",), (...,)]
        ```

    """

    _data: cached_property[TStatement]
    is_scalar: bool = internal.configurable_attribute(False)

    use_naive_filters: bool = internal.configurable_attribute(True)
    use_naive_search: bool = internal.configurable_attribute(True)
    session: AlchemySession = internal.configurable_attribute(
        default_factory=lambda self: model.Session,
    )

    EMPTY_STRING = object()

    def compute_data(self):
        stmt = self.get_base_statement()
        stmt = self.alter_statement(stmt)
        stmt = self.statement_with_filters(stmt)
        return self.statement_with_sorting(stmt)

    @abc.abstractmethod
    def get_base_statement(self) -> Any:
        """Return statement with minimal amount of columns and filters."""
        ...

    def compute_total(self, data: TStatement) -> int:
        return self.count_statement(data)

    def __iter__(self) -> Iterator[types.TData]:
        yield from self.execute_statement(self._data)

    def _execute(self, stmt: GenerativeSelect):
        return self.session.execute(stmt)

    def range(self, start: int, end: int) -> Iterable[types.TData]:
        stmt = self._data.limit(end - start).offset(start)
        return self.execute_statement(stmt)

    def execute_statement(self, stmt: TStatement) -> Iterable[types.TData]:
        result: Any
        if self.is_scalar:
            result = self.session.scalars(stmt)

        else:
            result = self.session.execute(stmt)

        return result

    def alter_statement(self, stmt: TStatement):
        """Add columns, joins and unconditional filters to statement."""
        return stmt

    def count_statement(self, stmt: TStatement) -> int:
        """Count number of items in query."""
        count_stmt: Select = sa.select(sa.func.count()).select_from(stmt.subquery())
        return cast(int, self._execute(count_stmt).scalar())

    def _into_clause(self, column: ColumnElement[Any], value: Any):
        if isinstance(value, list):
            return column.in_(value)

        if value is self.EMPTY_STRING:
            value = ""

        column_type = column.type.python_type
        if not isinstance(value, column_type) and value is not None:
            value = str(value)  # noqa: PLW2901
            column = sa.func.cast(column, sa.Text)

        return column == value

    def statement_with_filters(self, stmt: TStatement) -> TStatement:
        """Add normal filter to statement."""
        if not isinstance(stmt, Select):
            return stmt

        params = self.attached.params
        if self.use_naive_filters:
            filters = [
                name
                for name in self.attached.columns.filterable
                if name in params
                and params[name] != ""
                and name in stmt.selected_columns
            ]

            if filters:
                stmt = stmt.where(
                    sa.and_(
                        *[
                            self._into_clause(stmt.selected_columns[name], params[name])
                            for name in filters
                        ],
                    ),
                )

        if self.use_naive_search:
            if (q := params.get("q")) and "q" not in self.attached.columns.searchable:
                filters = [
                    name
                    for name in self.attached.columns.searchable
                    if name in stmt.selected_columns
                ]
                if filters:
                    stmt = stmt.where(
                        sa.or_(
                            *[
                                stmt.selected_columns[name].ilike(f"%{q}%")
                                for name in filters
                            ],
                        ),
                    )
            for name in self.attached.columns.searchable:
                if name not in params or name not in stmt.selected_columns:
                    continue
                value = params[name]
                stmt = stmt.where(stmt.selected_columns[name].ilike(f"%{value}%"))

        return stmt

    def statement_with_sorting(self, stmt: TStatement):
        """Specify sorting order for statement."""
        sort = self.attached.params.get("sort")
        if not sort:
            return stmt

        column, desc = internal.parse_sort(sort)

        if column not in self.attached.columns.sortable:
            log.warning("Unexpected sort value: %s", column)
            return stmt

        if column not in stmt.selected_columns:
            sort = column
            if desc:
                sort = f"{sort} DESC"
            return stmt.order_by(sa.text(sort))

        col_object = stmt.selected_columns[column]

        return stmt.order_by(col_object.desc() if desc else col_object.asc())


class TemporalSaData(BaseSaData[TStatement, types.TData, types.TDataCollection]):
    """Data source that supports pagination by datetime column.

    Attributes:
        temporal_column: column used for pagination

    Example:
        This class can be used as a base for SQL statement based data
        services. Collection that uses such data service must also use
        `pager.TemporalPager` instead of the default `pager.ClassicPager`.

        ```python
        import sqlalchemy as sa
        from datetime import date, timedelta
        from ckan import model

        class TemporalPackageData(data.TemporalSaData):
            def get_base_statement(self):
                return sa.select(model.Package.name, model.Package.metadata_created)
        ```

        ```pycon
        >>> col = collection.Collection(
        >>>     data_factory=TemporalPackageData,
        >>>     data_settings={"temporal_column": model.Package.metadata_created},
        >>>     pager_factory=pager.TemporalPager,
        >>>     pager_settings={"since": datetime.now() - timedelta(days=40)},
        >>> )
        >>> list(col)
        [("pkg1", datetime.datetime(2024, 6, 13, 10, 40, 22, 518511)), ...]
        ```

    """

    temporal_column: ColumnElement[Any] = internal.configurable_attribute()

    def range(self, start: date | None, end: date | None) -> Iterable[types.TData]:  # type: ignore
        stmt = self._data
        if not isinstance(stmt, Select):
            return self.execute_statement(stmt)

        if isinstance(self.temporal_column, str):
            col = stmt.selected_columns[self.temporal_column]
        else:
            col = self.temporal_column

        if start:
            stmt = stmt.where(col >= start)
        if end:
            stmt = stmt.where(col < end)

        return self.execute_statement(stmt)


class StatementSaData(BaseSaData[Select, types.TData, types.TDataCollection]):
    """Data source for arbitrary SQL statement.

    Warning:
        Final statement produced by `compute_data` and total number of results
        are cached. Call `refresh_data()` method of the service to rebuild the
        statement and refresh number of rows.

    Attributes:
        statement (sqlalchemy.sql.Select): select statement

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     data_factory=data.StatementSaData,
        >>>     data_settings={"statement": sa.select(model.User.name)},
        >>> )
        >>> list(col)
        [("default",), (...,)]
        ```

    """

    statement: Any = internal.configurable_attribute(None)

    def get_base_statement(self):
        """Return statement with minimal amount of columns and filters."""
        return self.statement


class UnionSaData(BaseSaData[Select, types.TData, types.TDataCollection]):
    """Data source for multiple SQL statement merged with UNION ALL.

    Warning:
        Final statement produced by `compute_data` and total number of results
        are cached. Call `refresh_data()` method of the service to rebuild the
        statement and refresh number of rows.

    Attributes:
        statements (sqlalchemy.sql.Select): select statements

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     data_factory=data.UnionSaData,
        >>>     data_settings={"statements": [
        >>>         sa.select(model.User.name, sa.literal("user")),
        >>>         sa.select(model.Package.name, sa.literal("package")),
        >>>         sa.select(model.Group.name, sa.literal("group")),
        >>>     ]},
        >>> )
        >>> list(col)
        [("default", "user"),
        ("warandpeace", "package"),
        ("my-cool-group", "group")]
        ```
    """

    statements: Iterable[GenerativeSelect] = internal.configurable_attribute(
        default_factory=lambda self: [],
    )

    def get_base_statement(self):
        """Return statement with minimal amount of columns and filters."""
        return sa.select(sa.union_all(*self.statements).subquery())


class ModelData(BaseSaData[Select, types.TData, types.TDataCollection]):
    """Data source for SQLAlchemy model.

    Warning:
        Final statement produced by `compute_data` and total number of results
        are cached. Call `refresh_data()` method of the service to rebuild the
        statement and refresh number of rows.

    Attributes:
        model: main model used by data source
        is_scalar: return model instance instead of collection of columns.
            /// details
                type: example

            Non-scalar collection returns rows as tuples of columns
            ```pycon
            >>> col = collection.Collection(
            >>>     data_factory=data.ModelData,
            >>>     data_settings={
            >>>         "model": model.User,
            >>>         "is_scalar": False,
            >>>     },
            >>> )
            >>> list(col)
            [("id-123-123", "user-name", ...), ...]
            ```

            Scalar collection yields model instances
            ```pycon
            >>> col = collection.Collection(
            >>>     data_factory=data.ModelData,
            >>>     data_settings={
            >>>         "model": model.User,
            >>>         "is_scalar": True,
            >>>     },
            >>> )
            >>> list(col)
            [<User id=id-123-123 name=user-name>, ...]
            ```
            ///
        static_columns: select only specified columns. If `is_scalar` flag
            enabled, only first columns from this list is returned.
            /// details
                type: example

            Non-scalar collection with `static_columns` produces tuples with values
            ```pycon
            >>> col = collection.Collection(
            >>>     data_factory=data.ModelData,
            >>>     data_settings={
            >>>         "model": model.User,
            >>>         "is_scalar": False,
            >>>         "static_columns": [model.User.name, model.User.sysadmin],
            >>>     },
            >>> )
            >>> list(col)
            [("default", True), ...]
            ```

            Scalar collection with `static_columns` yields values of the first column
            ```pycon
            >>> col = collection.Collection(
            >>>     data_factory=data.ModelData,
            >>>     data_settings={
            >>>         "model": model.User,
            >>>         "is_scalar": True,
            >>>         "static_columns": [model.User.name, model.User.sysadmin],
            >>>     },
            >>> )
            >>> list(col)
            ["default", ...]
            ```
            ///

        static_filters: apply filters to the select statement
            /// details
                type: example

            Filters are values produced by operations with model columns. The
            same thing you'd pass into `.filter`/`.where` methods.

            ```pycon
            >>> col = collection.Collection(
            >>>     data_factory=data.ModelData,
            >>>     data_settings={
            >>>         "model": model.User,
            >>>         "is_scalar": True,
            >>>         "static_filters": [model.User.sysadmin == True],
            >>>     },
            >>> )
            >>> list(col)
            [<User sysadmin=True ...>, ...]
            ```
            ///

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     data_factory=data.ModelData,
        >>>     data_settings={"model": model.User, "is_scalar": True},
        >>> )
        >>> list(col)
        [<User ...>, ...]
        ```

    """

    model: Any = internal.configurable_attribute(None)
    static_columns: list[sa.Column[Any] | Label[Any]] = internal.configurable_attribute(
        default_factory=lambda self: [],
    )
    static_filters: list[Any] = internal.configurable_attribute(
        default_factory=lambda self: [],
    )
    static_sources: dict[str, Any] = internal.configurable_attribute(
        default_factory=lambda self: {},
    )
    static_joins: list[tuple[str, Any, bool]] = internal.configurable_attribute(
        default_factory=lambda self: [],
    )

    def select_columns(self) -> Iterable[Any]:
        """Return list of columns for select statement."""
        if self.static_columns:
            return self.static_columns

        if not self.model:
            return []

        mapper = cast(Mapper, sa.inspect(self.model))
        return [self.model] if self.is_scalar else [mapper.columns]

    def get_extra_sources(self) -> dict[str, Any]:
        """Return mapping of additional models/subqueries for statement.

        Note: Don't call this method direclty. Instead, use `extra_sources`
        property, that caches return value of the current method. Extra sources
        must not change during query building process.

        Extra sources are used when the query contains subqueries or joins with
        aliased models. For example, in order to select column from joined
        subquery, instead of creating identical subqueries inside
        `select_columns` and `get_joins`, write the following code:

        >>> def get_extra_sources(self):
        >>>     return {"subq": make_subquery()}
        >>>
        >>> def select_columns(self):
        >>>     return [self.extra_sources["subq"].c.column_name]
        >>>
        >>> def get_joins(self):
        >>>     subq = self.extra_sources["subq"]
        >>>     return [("subq", subq.c.id == self.model.id, False)]

        """
        return self.static_sources

    @cached_property
    def extra_sources(self) -> dict[str, Any]:
        """Cache for extra sources.

        If you are going to override this property, don't forget to add
        `cached_property` decorator. The better option would be to override
        `get_extra_sources`. In this way you will not be affected if cache
        implementation changes in future.
        """
        return self.get_extra_sources()

    def get_joins(self) -> Iterable[tuple[str, Any, bool]]:
        """Return join details for extra_sources.

        Every item is a tuple of of (extra_source_name, join_condition,
        outerjoin_flag)
        """
        return self.static_joins

    def apply_joins(self, stmt: Select) -> Select:
        """Return list of columns for select statement."""
        sources = self.extra_sources

        for name, condition, isouter in self.get_joins():
            stmt = stmt.join(sources[name], condition, isouter=isouter)

        return stmt

    def get_base_statement(self):
        """Return statement with minimal amount of columns and filters."""
        columns = self.select_columns()
        stmt = sa.select(*columns)
        if self.model:
            stmt = stmt.select_from(self.model)

        return self.apply_joins(stmt)

    def statement_with_filters(self, stmt: Select):
        """Add normal filter to statement."""
        for cond in self.static_filters:
            stmt = stmt.where(cond)

        return super().statement_with_filters(stmt)



from __future__ import annotations

from typing import Any, Iterable

import ckan.plugins.toolkit as tk
from ckan.authz import is_authorized_boolean
from ckan.logic import parse_params

from ckanext.collection import internal, types
from ckanext.collection.utils.columns import TableColumns
from ckanext.collection.utils.data import Data
from ckanext.collection.utils.data.db import TableData
from ckanext.collection.utils.filters import DbFilters, Filters, TableFilters
from ckanext.collection.utils.serialize import HtmlSerializer, HtmxTableSerializer

from .base import Collection
from .db import DbCollection


class ExplorerSerializer(HtmlSerializer[types.TDataCollection]):
    extend_page_template: bool = internal.configurable_attribute(
        default_factory=lambda self: bool(
            tk.request and not tk.request.headers.get("hx-request"),
        ),
    )
    main_template: str = internal.configurable_attribute(
        "collection/serialize/explorer/main.html",
    )


class CollectionExplorer(Collection):
    """Collection of all registered collections.

    It exists for debugging and serves as an example of non-canonical usage of
    collections.
    """

    SerializerFactory = ExplorerSerializer.with_attributes(
        main_template="collection/serialize/collection_explorer/main.html",
    )

    class DataFactory(Data[Any, "CollectionExplorer"]):
        def compute_data(self) -> Iterable[Any]:
            params = parse_params(tk.request.args) if tk.request else {}
            names = self.attached.params.get("collection")
            if not names:
                return []

            if isinstance(names, str):
                names = [names]

            return [
                internal.get_collection(
                    name,
                    params,
                    serializer_settings={"extend_page_template": False},
                )
                for name in names
            ]

    class FiltersFactory(Filters["CollectionExplorer"], internal.UserTrait):
        static_collections: Iterable[str] = internal.configurable_attribute(
            default_factory=lambda self: list(
                map(str, internal.collection_registry.members),
            ),
        )

        def make_filters(self) -> Iterable[types.Filter[Any]]:
            return [
                types.SelectFilter(
                    name="collection",
                    type="select",
                    options={
                        "label": "Collection",
                        "options": [types.SelectOption(value="", text="")]
                        + [
                            types.SelectOption(value=name, text=name)
                            for name in self.static_collections
                            if name != self.attached.name
                            and is_authorized_boolean(
                                "collection_view_render",
                                {"user": self.user},
                                {"name": name},
                            )
                        ],
                    },
                ),
            ]


class DbExplorer(DbCollection):
    SerializerFactory = ExplorerSerializer.with_attributes(
        main_template="collection/serialize/db_explorer/main.html",
    )

    class DataFactory(Data[Any, "DbExplorer"]):
        def compute_data(self) -> Iterable[Any]:
            params = parse_params(tk.request.args)
            tables = self.attached.params.get("table")
            if not tables:
                return []

            if isinstance(tables, str):
                tables = [tables]

            _hidden = self.attached.params.get("hidden", "").split(",")
            hidden: set[str] = set(filter(None, map(str.strip, _hidden)))

            _visible = self.attached.params.get("visible", "").split(",")
            visible: set[str] = set(filter(None, map(str.strip, _visible))) - hidden

            _allowed_filters = self.attached.params.get("allowed_filters", "").split(
                ",",
            )
            allowed_filters: set[str] = set(
                filter(None, map(str.strip, _allowed_filters)),
            )

            _searchable_fields = self.attached.params.get(
                "searchable_fields",
                "",
            ).split(
                ",",
            )
            searchable_fields: set[str] = (
                set(filter(None, map(str.strip, _searchable_fields))) - allowed_filters
            )

            return [
                DbCollection(
                    name,
                    params,
                    db_connection_factory=self.attached.DbConnectionFactory,
                    data_factory=TableData.with_attributes(
                        table=name,
                        use_naive_filters=True,
                        use_naive_search=True,
                    ),
                    columns_factory=TableColumns.with_attributes(
                        table=name,
                        visible=visible or TableColumns.Default.NOT_HIDDEN,
                        hidden=hidden,
                        filterable=allowed_filters,
                        searchable=searchable_fields,
                    ),
                    filters_factory=TableFilters.with_attributes(table=name),
                    serializer_factory=HtmxTableSerializer,
                    serializer_settings={
                        "render_url": tk.h.url_for(
                            "ckanext-collection.render",
                            name=self.attached.name,
                            **{
                                k: v
                                for k, v in params.items()
                                if not k.startswith(f"{name}:")
                            },
                        ),
                    },
                )
                for name in tables
            ]

    class FiltersFactory(DbFilters["DbExplorer"]):
        static_tables: Iterable[str] = internal.configurable_attribute(
            default_factory=lambda self: [],
        )

        def make_filters(self) -> Iterable[types.Filter[Any]]:
            tables = self.static_tables
            if not tables:
                tables = self.attached.db_connection.inspector.get_table_names()

            return [
                types.SelectFilter(
                    name="table",
                    type="select",
                    options={
                        "label": "Table",
                        "options": [types.SelectOption(value="", text="")]
                        + [
                            types.SelectOption(value=name, text=name) for name in tables
                        ],
                    },
                ),
                types.InputFilter(
                    name="visible",
                    type="input",
                    options={
                        "label": "Visible columns(leave empty to show all columns)",
                        "placeholder": "comma-separated",
                    },
                ),
                types.InputFilter(
                    name="hidden",
                    type="input",
                    options={
                        "label": "Hidden columns",
                        "placeholder": "comma-separated",
                    },
                ),
                types.InputFilter(
                    name="allowed_filters",
                    type="input",
                    options={
                        "label": "Filterable columns(by exact values)",
                        "placeholder": "comma-separated",
                    },
                ),
                types.InputFilter(
                    name="searchable_fields",
                    type="input",
                    options={
                        "label": "Searchable columns(by case-insensitive fragment)",
                        "placeholder": "comma-separated",
                    },
                ),
            ]



from __future__ import annotations

from typing import Any

from typing_extensions import Self

from ckanext.collection import types
from ckanext.collection.utils.columns import DbColumns
from ckanext.collection.utils.data import DbData
from ckanext.collection.utils.db_connection import DbConnection
from ckanext.collection.utils.filters import DbFilters

from .base import Collection


class DbCollection(Collection, types.BaseDbCollection):
    _service_names: tuple[str, ...] = ("db_connection",) + Collection._service_names
    DbConnectionFactory: type[DbConnection[Self]] = DbConnection
    DataFactory = DbData
    ColumnsFactory = DbColumns
    FiltersFactory = DbFilters

    def make_db_connection(self, **kwargs: Any) -> DbConnection[Self]:
        """Return connection."""
        return self.DbConnectionFactory(self, **kwargs)



from __future__ import annotations

from functools import cached_property
from typing import Any, Generic, Iterable, Iterator

from ckanext.collection import internal, types


class Data(
    types.BaseData,
    internal.Domain[types.TDataCollection],
    Generic[types.TData, types.TDataCollection],
):
    """Base data source for collection.

    This class defines an outline of the data service. In basic case, sublcass
    should override `compute_data` method and return a Sequence from it to keep
    all methods functional.

    Example:
        ```python
        class MyData(data.Data):
            def compute_data(self):
                return range(1, 20)
        ```

    """

    @cached_property
    def _data(self):
        return self.compute_data()

    @cached_property
    def _total(self) -> int:
        return self.compute_total(self._data)

    def __iter__(self) -> Iterator[types.TData]:
        yield from self._data

    def __len__(self):
        return self.total

    def refresh_data(self):
        """Update data and counters.

        Use this method when parameters that affect Data are changed. This is a
        cheaper alternative of recreating the whole collection with new
        parameters.

        """
        self._data = self.compute_data()
        self._total = self.compute_total(self._data)

    def __getitem__(self, key: Any):
        if isinstance(key, slice):
            return self.range(key.start, key.stop)
        return self.at(key)

    def compute_data(self) -> Any:
        """Produce data."""
        return []

    def compute_total(self, data: Any) -> int:
        """Compute number of data records."""
        return len(data)

    def range(self, start: Any, end: Any) -> Iterable[types.TData]:
        """Slice data."""
        return self._data[start:end]

    def at(self, index: Any) -> types.TData:
        """Slice data."""
        return self._data[index]

    @property
    def total(self):
        return self._total



from __future__ import annotations

from . import config


def collection_include_htmx_asset() -> bool:
    """Check if HTMX needs to be added to webassets."""
    return config.include_htmx_asset()


def collection_init_modules_in_htmx_response():
    """Check if HTMX needs to be initialize CKAN modules for every HTMX response."""
    return config.htmx_init_modules()



"""Interface."""

from __future__ import annotations

from ckan.plugins import Interface

from ckanext.collection.types import CollectionFactory


class ICollection(Interface):
    """Extend functionality of ckanext-collections.

    Example:
        ```python
        import ckan.plugins as p
        from ckanext.collection import shared

        class MyPlugin(p.SingletonPlugin):
            p.implements(shared.ICollection, inherit=True)

            def get_collection_factories(self) -> dict[str, CollectionFactory]:
                return {...}
        ```
    """

    def get_collection_factories(self) -> dict[str, CollectionFactory]:
        """Register named collection factories.

        Example:
            ```python
            def get_collection_factories(self) -> dict[str, CollectionFactory]:
                return {
                    "packages": PackageCollection,
                }
            ```

        Returns:
            mapping of global collection name to collection factory


        """
        return {}



import pytest

from ckanext.collection import internal


@pytest.fixture()
def collection_registry():
    """Collection registry cleaned after each test."""
    yield internal.collection_registry
    internal.collection_registry.reset()



from __future__ import annotations

from datetime import date, timedelta
from typing import Any

from ckanext.collection import internal, types


class Pager(types.BasePager, internal.Domain[types.TDataCollection]):
    """Base for pager service.

    This class must be extended by every implementation of the pager service.

    Example:
        ```python
        class StaticPager(Pager):
            @property
            def start(self) -> int:
                return 0

            @property
            def end(self) -> int:
                return 10

            @property
            def size(self) -> int:
                return 10
        ```

    """


class ClassicPager(Pager[types.TDataCollection]):
    """Page-number based pagination.

    Attributes:
        page: current active page
        rows_per_page: max number of items per page
        prioritize_params: if `params` contain pagination details,
            use them instead of `pager_settings`. Defaults to True.

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     data_factory=data.StaticData,
        >>>     data_settings={"data": range(1, 100)},
        >>>     pager_factory=pager.ClassicPager,
        >>>     pager_settings={"page": 2, "rows_per_page": 5},
        >>> )
        >>> list(col)
        [6, 7, 8, 9, 10]
        ```
    """

    prioritize_params: bool = internal.configurable_attribute(True)
    page: int = internal.configurable_attribute(1)
    rows_per_page: int = internal.configurable_attribute(10)

    def __init__(self, obj: types.TDataCollection, /, **kwargs: Any):
        """Use `page` and `rows_per_page` parameters."""
        super().__init__(obj, **kwargs)

        if self.prioritize_params:
            self.page = max(int(self.attached.params.get("page", self.page)), 1)

            self.rows_per_page = max(
                int(
                    self.attached.params.get(
                        "rows_per_page",
                        self.rows_per_page,
                    ),
                ),
                0,
            )

    @property
    def start(self) -> int:
        return self.size * self.page - self.size

    @property
    def end(self) -> int:
        return self.start + self.size

    @property
    def size(self) -> Any:
        return self.rows_per_page


class OffsetPager(Pager[types.TDataCollection]):
    """Limit/offset based pagination.

    Attributes:
        offset: number of items to skip
        limit: max number of items per page

    Example:
        ```pycon
        >>> col = collection.Collection(
        >>>     data_factory=data.StaticData,
        >>>     data_settings={"data": range(1, 100)},
        >>>     pager_factory=pager.OffsetPager,
        >>>     pager_settings={"offset": 2, "limit": 3},
        >>> )
        >>> list(col)
        [3, 4, 5]
        ```
    """

    offset: int = internal.configurable_attribute(0)
    limit: int = internal.configurable_attribute(10)

    @property
    def start(self) -> int:
        return self.offset

    @property
    def end(self) -> int:
        return self.start + self.size

    @property
    def size(self) -> Any:
        return self.limit


class TemporalPager(Pager[types.TDataCollection]):
    """Date based pagination.

    Data service of the collection that uses TemporalPager must implement
    `range` as `range(start: date | None, end: date | None)`.

    Attributes:
        since: date of the oldest record(`>=`)
        until: date of the newest record(`<`)

    Example:
        Define data service that supports date ranges.

        ```python
        from datetime import datetime, timedelta

        class TemporalModelData(data.TemporalSaData, data.ModelData):
            pass
        ```

        Initialize the collection

        ```python
        >>> col = collection.Collection(
        >>>     data_factory=TemporalModelData,
        >>>     data_settings={
        >>>         "model": model.Package,
        >>>         "temporal_column": model.Package.metadata_created,
        >>>     },
        >>>     pager_factory=pager.TemporalPager,
        >>>     pager_settings={"since": datetime.now() - timedelta(days=1)},
        >>> )
        >>> list(col)
        [(...package created yesterday), ...]
        ```

    """

    since: date | None = internal.configurable_attribute(None)
    until: date | None = internal.configurable_attribute(None)

    @property
    def start(self) -> date | None:
        return self.since

    @property
    def end(self) -> date | None:
        return self.until

    @property
    def size(self) -> timedelta:
        return (self.end or date.max) - (self.start or date.min)



from __future__ import annotations

import pytest

from ckanext.collection.utils import Collection, filters


@pytest.fixture()
def collection():
    return Collection("", {})


class TestFilters:
    def test_settings(self, collection: Collection):
        obj = filters.Filters(collection)
        assert obj.filters == []
        assert obj.actions == []

        obj = filters.Filters(collection, static_filters=[1, 2, 3])
        assert obj.filters == [1, 2, 3]
        assert obj.actions == []



from __future__ import annotations

import enum
from typing import Any, cast

from ckanext.collection import internal, types


class Columns(
    types.BaseColumns,
    internal.Domain[types.TDataCollection],
):
    """Collection of columns details for filtering/rendering.

    Attributes:
        names: list of all available columns
        visible: columns that can be viewed
        sortable: columns that can be sorted
        filterable: columns that can be filtered using exect match
        searchable: columns that can be searched by partial match
        labels: UI labels for columns
    """

    class Default(internal.Sentinel, enum.Enum):
        ALL = enum.auto()
        NOT_HIDDEN = enum.auto()
        NONE = enum.auto()

    names: list[str] = internal.configurable_attribute(default_factory=lambda self: [])
    hidden: set[str] = internal.configurable_attribute(
        default_factory=lambda self: set(),
    )
    visible: set[str] = internal.configurable_attribute(Default.NOT_HIDDEN)
    sortable: set[str] = internal.configurable_attribute(Default.NONE)
    filterable: set[str] = internal.configurable_attribute(Default.NONE)
    searchable: set[str] = internal.configurable_attribute(Default.NONE)
    labels: dict[str, str] = internal.configurable_attribute(Default.ALL)

    serializers: dict[
        str,
        list[tuple[str, dict[str, Any]]],
    ] = internal.configurable_attribute(
        default_factory=lambda self: {},
    )

    def __init__(self, obj: types.TDataCollection, **kwargs: Any):
        super().__init__(obj, **kwargs)
        self.configure_attributes()

    def configure_attributes(self):
        if self.hidden is self.Default.ALL:
            self.hidden = set(self.names)
        elif isinstance(self.hidden, self.Default):
            self.hidden = set()

        self.visible = self._compute_set(self.visible)

        if not self.hidden:
            self.hidden = set(self.names) - self.visible

        self.sortable = self._compute_set(self.sortable)
        self.filterable = self._compute_set(self.filterable)
        self.searchable = self._compute_set(self.searchable)

        if isinstance(self.labels, self.Default):
            self.labels = {c: c for c in self._compute_set(self.labels)}

    def _compute_set(self, value: Default | set[str]):
        if value is self.Default.NONE:
            return cast("set[str]", set())

        if value is self.Default.ALL:
            return set(self.names)

        if value is self.Default.NOT_HIDDEN:
            return {c for c in self.names if c not in self.hidden}

        return value

    def get_primary_order(self, name: str) -> str:
        """Format column name as a primary(asc) order value.

        Examples:
            Hehe haha

            >>> col = Collection()
            >>> col.columns.get_primary_order("name")
            "name"

        Args:
            name: the name of the sorted column

        Returns:
            column name adapted for primary(ascending) sorting.
        """
        return name

    def get_secondary_order(self, name: str) -> str:
        """Format column name as a secondary(desc) order value.

        Examples:
            >>> col = Collection()
            >>> col.columns.get_secondary_order("name")
            "-name"


        Args:
            name: the name of the sorted column

        Returns:
            column name adapted for secondary(descending) sorting.

        """
        return f"-{name}"


class DbColumns(Columns[types.TDbCollection]):
    pass


class TableColumns(DbColumns[types.TDbCollection]):
    table: str = internal.configurable_attribute()
    filterable: set[str] = internal.configurable_attribute(
        default_factory=lambda self: self.Default.NONE,
    )

    def configure_attributes(self):
        self.names = [
            c["name"]
            for c in self.attached.db_connection.inspector.get_columns(self.table)
        ]
        super().configure_attributes()



from __future__ import annotations

import pytest

from ckan import model

from ckanext.collection import internal
from ckanext.collection.utils import *


class TestOverview:
    def test_static(self):
        col = StaticCollection("name", {})
        assert list(col) == []

        col = StaticCollection("name", {}, data_settings={"data": [1, 2, 3]})
        assert list(col) == [1, 2, 3]

    @pytest.mark.usefixtures("clean_db")
    def test_model(self):
        col = ModelCollection(
            "",
            {},
            data_settings={"is_scalar": True, "model": model.User},
        )

        for user in col:
            assert isinstance(user, model.User)

    @pytest.mark.usefixtures("clean_db", "clean_index", "package")
    def test_api_search(self):
        col = ApiSearchCollection("", {}, data_settings={"action": "package_search"})

        for pkg in col:
            assert isinstance(pkg, dict)

    @pytest.mark.usefixtures("clean_db", "user")
    def test_api(self):
        col = ApiCollection("", {}, data_settings={"action": "user_list"})

        for user in col:
            assert isinstance(user, dict)


class TestInitialization:
    def test_params(self):
        col = Collection("hello", {"hello:a": 1, "b": 2, "world:c": 3})
        assert col.params == {"a": 1}

        col = Collection("world", {"hello:a": 1, "b": 2, "world:c": 3})
        assert col.params == {"c": 3}

    def test_multi(self):
        params = {"users:page": 2, "packages:page": 5}

        users = ModelCollection("users", params, data_settings={"model": model.User})
        packages = ModelCollection(
            "packages",
            params,
            data_settings={"model": model.Package},
        )

        assert isinstance(users.pager, ClassicPager)
        assert isinstance(packages.pager, ClassicPager)

        assert users.pager.page == 2
        assert packages.pager.page == 5


class TestServices:
    def test_default_initialization(self):
        col = Collection("name", {})
        assert isinstance(col.data, Data)
        assert col.pager
        assert col.serializer
        assert col.columns
        assert col.filters

    def test_service_replacement(self):
        col = Collection("name", {})
        static_data = StaticData(col, data=[1, 2, 3])
        col.replace_service(static_data)

        assert list(col) == [1, 2, 3]

    def test_factory(self):
        col = Collection("name", {}, data_factory=StaticData)
        assert list(col) == []

        col.data.data = [1, 2, 3]
        col.data.refresh_data()
        assert list(col) == [1, 2, 3]

    def test_custom_settings(self):
        col = Collection(
            "name",
            {},
            data_factory=StaticData,
            data_settings={"data": [1, 2, 3]},
        )
        assert list(col) == [1, 2, 3]

    def test_custom_service_factory(self):
        class MyCollection(Collection):
            DataFactory = StaticData

        col = MyCollection("name", {}, data_settings={"data": [1, 2, 3]})
        assert list(col) == [1, 2, 3]

    def test_with_attributes(self):
        class MyCollection(Collection):
            DataFactory = StaticData.with_attributes(data=[1, 2, 3])

        col = MyCollection("name", {})
        assert list(col) == [1, 2, 3]


class TestCommonLogic:
    def test_attached(self):
        col = Collection("name", {})
        assert col.data.attached is col
        assert col.pager.attached is col
        assert col.columns.attached is col

        another_col = Collection("another-name", {}, data_instance=col.data)
        assert col.data.attached is not col
        assert col.data.attached is another_col
        assert col.data is another_col.data

    def test_settings(self):
        col = Collection("name", {})
        data = StaticData(col, data=[], not_real=True)
        assert hasattr(data, "data")
        assert not hasattr(data, "not_real")

    def test_configurable_attributes(self):
        col = Collection("name", {})

        class MyData(StaticData):
            i_am_real = internal.configurable_attribute(False)

        data = MyData(col, data=[], i_am_real=True)
        assert hasattr(data, "data")
        assert hasattr(data, "i_am_real")
        assert data.i_am_real is True

    def test_configurable_attribute_default_factory(self):
        col = Collection("name", {})

        class MyData(StaticData):
            ref = 42
            i_am_real = internal.configurable_attribute(
                default_factory=lambda self: self.ref * 10,
            )

        data = MyData(col, data=[])
        assert data.i_am_real == 420



from __future__ import annotations

import abc
from collections.abc import Sized
from typing import Any, Callable, Generic, Iterable, Literal, Protocol, Union, TypeVar

import sqlalchemy as sa
from sqlalchemy.engine import Engine
from typing_extensions import NotRequired, TypeAlias, TypedDict


# CollectionFactory: TypeAlias = "Callable[[str, dict[str, Any]], BaseCollection[Any]]"
class CollectionFactory(Protocol):
    def __call__(
        self,
        name: str,
        params: dict[str, Any],
        /,
        **kwargs: Any,
    ) -> BaseCollection:
        ...


TCollection = TypeVar("TCollection", bound="BaseCollection")
TDataCollection = TypeVar("TDataCollection", bound="BaseCollection")
TDbCollection = TypeVar("TDbCollection", bound="BaseDbCollection")
TFilterOptions = TypeVar("TFilterOptions")
TSerialized = TypeVar("TSerialized")
TData = TypeVar("TData")


class Service:
    """Marker for service classes used by collection."""

    def __init__(self, obj: Any, /, **kwargs: Any):
        pass

    @property
    @abc.abstractmethod
    def service_name(self) -> str:
        """Name of the service instance used by collection."""
        ...


class BaseColumns(abc.ABC, Service):
    """Declaration of columns properties."""

    names: list[str]
    visible: set[str]
    sortable: set[str]
    filterable: set[str]
    searchable: set[str]
    labels: dict[str, str]
    serializers: dict[str, list[tuple[str | ValueSerializer, dict[str, Any]]]]

    @property
    def service_name(self):
        return "columns"


class BaseData(abc.ABC, Sized, Iterable[Any], Service):
    """Declaration of data properties."""

    @property
    @abc.abstractmethod
    def total(self) -> int:
        """Total number of data records."""
        ...

    @abc.abstractmethod
    def range(self, start: Any, end: Any) -> Iterable[Any]:
        """Slice data using specified limits."""
        ...

    @property
    def service_name(self):
        return "data"


class BasePager(abc.ABC, Service):
    """Declaration of pager properties."""

    params: dict[str, Any]

    @property
    @abc.abstractmethod
    def size(self) -> Any:
        """Range of the pager.

        In classic pager it may be the number of items per page. For date range
        pager, it can be a timespan within which we are searching for records.

        """
        ...

    @property
    @abc.abstractmethod
    def start(self) -> Any:
        """Inclusive lower bound of the page.

        For classic limit/offset pagination, start:0 means that index of the
        first element is 0.

        """
        ...

    @property
    @abc.abstractmethod
    def end(self) -> Any:
        """Exclusive upper bound of the page.

        For classic limit/offset pagination, end:10 means that index of the
        last element is less than 10.

        """
        ...

    @property
    def service_name(self):
        return "pager"


class BaseSerializer(abc.ABC, Service):
    @abc.abstractmethod
    def serialize(self) -> Any:
        ...

    @property
    def service_name(self):
        return "serializer"

    @abc.abstractmethod
    def serialize_value(self, value: Any, name: str, record: Any):
        """Transform record's value into its serialized form."""
        ...

    @abc.abstractmethod
    def dictize_row(self, row: Any) -> dict[str, Any]:
        """Transform single data record into serializable dictionary."""
        ...


class BaseFilters(abc.ABC, Service):
    """Declaration of filters properties."""

    filters: Iterable[Filter[Any]]
    actions: Iterable[Filter[Any]]

    @property
    def service_name(self):
        return "filters"


class BaseCollection(abc.ABC, Iterable[Any]):
    """Declaration of collection properties."""

    name: str
    params: dict[str, Any]

    columns: BaseColumns
    data: BaseData
    filters: BaseFilters
    pager: BasePager
    serializer: BaseSerializer


class BaseDbConnection(abc.ABC, Service):
    engine: Engine

    @property
    def service_name(self):
        return "db_connection"

    @property
    def inspector(self):
        return sa.inspect(self.engine)


class BaseDbCollection(BaseCollection):
    db_connection: BaseDbConnection


class Filter(TypedDict, Generic[TFilterOptions]):
    """Filter details."""

    name: str
    type: Any
    options: TFilterOptions


class SelectOption(TypedDict):
    text: str
    value: str


class SelectFilterOptions(TypedDict):
    label: str
    options: Iterable[SelectOption]


class InputFilterOptions(TypedDict):
    label: str
    placeholder: NotRequired[str]
    type: NotRequired[str]


class ButtonFilterOptions(TypedDict):
    label: str
    type: NotRequired[str]
    attrs: NotRequired[dict[str, Any]]


class StaticLinkFilterOptions(TypedDict):
    label: str
    href: str
    attrs: NotRequired[dict[str, Any]]


class DynamicLinkFilterOptions(TypedDict):
    label: str
    endpoint: str
    kwargs: dict[str, Any]
    attrs: NotRequired[dict[str, Any]]


class SelectFilter(Filter[SelectFilterOptions]):
    type: Literal["select"]


class InputFilter(Filter[InputFilterOptions]):
    type: Literal["input"]


class ButtonFilter(Filter[ButtonFilterOptions]):
    type: Literal["button"]


class LinkFilter(Filter[Union[StaticLinkFilterOptions, DynamicLinkFilterOptions]]):
    type: Literal["link"]


ValueSerializer: TypeAlias = Callable[
    [Any, "dict[str, Any]", str, Any, BaseSerializer],
    Any,
]



from __future__ import annotations

from ckanext.collection.utils.data import StaticData

from .api import ApiCollection, ApiSearchCollection
from .base import Collection
from .db import DbCollection
from .explorer import CollectionExplorer, DbExplorer
from .model import ModelCollection

__all__ = [
    "Collection",
    "DbCollection",
    "ApiCollection",
    "ApiSearchCollection",
    "ModelCollection",
    "CollectionExplorer",
    "DbExplorer",
    "StaticCollection",
]


class StaticCollection(Collection):
    DataFactory = StaticData
